from __future__ import annotations

from typing import Optional, Tuple

from jax import numpy as jnp

from tjax import Generator, Tensor, real_dtype

__all__ = ['leaky_integrate', 'diffused_leaky_integrate']


def leaky_integrate(value: Tensor,
                    time_step: real_dtype,
                    drift: Optional[Tensor] = None,
                    decay: Optional[Tensor] = None,
                    *,
                    leaky_average: bool = False) -> Tensor:
    """
    Update the value so that it is the leaky integral (or leaky average).
    Args:
        value: The current value of the leaky integral or average.
        time_step: The number of seconds that have passed.
        decay: If provided, must be positive, and the value decays by exp(-decay) every second.
        drift: If provided, the value increases by this every second.
        leaky_average: A flag indicating a leaky average rather than a leaky integral.  This scales
            the drift by the real component (in case the decay is complex) of the decay.
    """
    if drift is None:
        if decay is None:
            return value
        return value * jnp.exp(-decay * time_step)

    if decay is None:
        if leaky_average:
            raise ValueError
        return value + drift * time_step

    scaled_integrand = (drift / decay) * -jnp.expm1(-decay * time_step)

    if leaky_average:
        scaled_integrand *= decay.real

    return value * jnp.exp(-decay * time_step) + scaled_integrand


def diffused_leaky_integrate(value: Tensor,
                             time_step: real_dtype,
                             rng: Generator,
                             diffusion: Tensor,
                             drift: Optional[Tensor] = None,
                             decay: Optional[Tensor] = None,
                             *,
                             leaky_average: bool = False) -> Tuple[Tensor, Generator]:
    """
    Update an Ornstein-Uhlenbeck process.

    Args:
        value: The current value of the leaky integral or average.
        time_step: The number of seconds that have passed.
        rng: The random number generator for the stochastic process.
        diffusion: The diffusion for the stochastic process.
        decay: If provided, the value decays by exp(-decay) every second.
        drift: If provided, the value increases by this every second.
        leaky_average: A flag indicating a leaky average rather than a leaky integral.  This scales
            the drift by the real component (in case the decay is complex) of the decay.
    """
    variance = (diffusion * time_step
                if decay is None
                else diffusion / decay * -jnp.expm1(-decay * time_step))
    new_rng, jump = rng.normal(jnp.sqrt(variance), shape=value.shape)
    return (leaky_integrate(value, time_step, drift, decay, leaky_average=leaky_average) + jump,
            new_rng)



from typing import Generic, TypeVar

from tjax import PyTree

from .augmented import State
from .combinator import IteratedFunctionWithCombinator
from .comparing import ComparingIteratedFunction, ComparingState
from .iterated_function import Parameters

__all__ = ['ComparingIteratedFunctionWithCombinator']


Comparand = TypeVar('Comparand', bound=PyTree)


class ComparingIteratedFunctionWithCombinator(
        IteratedFunctionWithCombinator[Parameters, State, ComparingState[State, Comparand]],
        ComparingIteratedFunction[Parameters, State, Comparand],
        Generic[Parameters, State, Comparand]):
    pass



from typing import Any, Hashable, List, MutableMapping, Sequence, Tuple, Type, TypeVar

import cooperative_dataclasses as dataclasses
from cooperative_dataclasses import MISSING, Field, FrozenInstanceError, InitVar
from jax.tree_util import register_pytree_node

from .annotations import PyTree
from .display import display_class, display_key_and_value

__all__ = ['dataclass', 'field', 'Field', 'FrozenInstanceError', 'InitVar', 'MISSING']


T = TypeVar('T', bound=Any)


def dataclass(clz: Type[T]) -> Type[T]:
    """
    Returns the same class as was passed in, with dunder methods added based on the fields defined
    in the class.

    Examines PEP 526 annotations to determine fields.  Default values for fields are provided using
    assignment.  To mark fields as JAX static fields rather than JAX pytree fields, use the `field`
    function.

    For example::
    ```python
    from __future__ import annotations

    from typing import ClassVar

    from tjax import dataclass, field, Tensor
    from jax import numpy as jnp
    from jax import grad

    @dataclass
    class LearnedParameter:
        weight: Tensor
        constrain_positive: bool = field(pytree_like=False)
        minimum_positive_weight: ClassVar[Tensor] = 1e-6

        def trained(self,
                    self_bar: LearnedParameter,
                    learning_rate: float) -> LearnedParameter:
            weight_bar = self_bar.weight
            weight = self.weight - weight_bar * learning_rate
            if self.constrain_positive:
                weight = jnp.maximum(weight, self.minimum_positive_weight)
            return LearnedParameter(weight=weight,
                                    constrain_positive=self.constrain_positive)

    def loss(w: LearnedParameter) -> float:
        return jnp.square(w.weight - 3.3)

    w = LearnedParameter(2.0, True)
    w_bar = grad(loss)(w)
    new_w = w.trained(w_bar, 1e-4)
    ```

    `dataclass` includes a convenient replace method::

        w.replace(weight=3.4)

    Since this dataclass is a pytree, all of JAX's functions that accept pytrees work with it,
    including iteration, differentiation, and `jax.tree_util` functions.

    Another benefit is the display of dataclasses.  `print(new_w)` gives::
    ```
    LearnedParameter
        weight=Jax Array ()
                2.0003
        constrain_positive=True
    ```
    """
    # pylint: disable=protected-access

    # Apply dataclass function to clz.
    data_clz: Type[T] = dataclasses.dataclass(frozen=True)(clz)  # type: ignore

    # Partition fields into hashed, tree, and uninitialized.
    hashed_fields: List[str] = []
    tree_fields: List[str] = []

    for field_info in dataclasses.fields(data_clz):  # type: ignore
        if not field_info.init:
            continue
        if field_info.metadata.get('pytree_like', True):
            tree_fields.append(field_info.name)
        else:
            hashed_fields.append(field_info.name)

    # Generate additional methods.
    def __repr__(self: T) -> str:
        return str(self.display())

    def display(self: T, show_values: bool = True, indent: int = 0) -> str:
        retval = display_class(type(self))
        for field_info in dataclasses.fields(data_clz):  # type: ignore
            retval += display_key_and_value(
                field_info.name, getattr(self, field_info.name), "=", show_values, indent)
        return retval

    def tree_flatten(x: T) -> Tuple[Sequence[PyTree], Hashable]:
        hashed = tuple(getattr(x, name) for name in hashed_fields)
        trees = tuple(getattr(x, name) for name in tree_fields)
        return trees, hashed

    def tree_unflatten(cls: Type[T], hashed: Hashable, trees: Sequence[PyTree]) -> T:
        if not isinstance(hashed, tuple):
            raise TypeError
        hashed_args = dict(zip(hashed_fields, hashed))
        tree_args = dict(zip(tree_fields, trees))
        return cls(**hashed_args, **tree_args)

    # Assign methods to the class.
    data_clz.__repr__ = __repr__  # type: ignore
    data_clz.display = display  # type: ignore
    data_clz.tree_flatten = tree_flatten  # type: ignore
    data_clz.tree_unflatten = classmethod(tree_unflatten)  # type: ignore

    # Assign field lists to the class.
    data_clz.tree_fields = tree_fields  # type: ignore
    data_clz.hashed_fields = hashed_fields  # type: ignore

    # Register the class as a JAX PyTree.
    register_pytree_node(data_clz, tree_flatten, data_clz.tree_unflatten)  # type: ignore

    return data_clz


def field(pytree_like: bool = True, **kwargs: Any) -> dataclasses.Field:
    """
    Args:
        pytree_like: Indicates whether a field is a pytree or static.  Pytree fields are
            differentiated and traced.
        kwargs: Any of the keyword arguments from `dataclasses.field`.
    """
    return dataclasses.field(metadata={**kwargs.pop('metadata', {}),
                                       'pytree_like': pytree_like},
                             **kwargs)


def document_dataclass(pdoc: MutableMapping[str, Any], name: str) -> None:
    pdoc[f'{name}.hashed_fields'] = False
    pdoc[f'{name}.tree_fields'] = False
    pdoc[f'{name}.tree_flatten'] = False
    pdoc[f'{name}.tree_unflatten'] = False
    pdoc[f'{name}.display'] = False
    pdoc[f'{name}.replace'] = False



from typing import Tuple, Type, TypeVar

from jax import numpy as jnp

from .annotations import Tensor
from .dataclass import dataclass

__all__ = ['LogCotangent']

T = TypeVar('T', bound='LogCotangent')


@dataclass
class LogCotangent:
    """
    LogCotangent logs cotangents in a differentiated jax function.  For example:

        from jax import grad
        from jax import numpy as jnp

        from tjax import LogCotangent

        def loss(x, w, log_cotangent):
            y = x * w
            z = log_cotangent.forward(y)
            return jnp.sum(jnp.square(2.0 - z))

        x = jnp.array([1.0, 2.0])
        w = jnp.array([2.2, 3.5])
        lg_bar = grad(loss, 2)(x, w, LogCotangent.create(shape=x.shape))

        # lg_bar.cotangent now holds the transmitted cotangent.
    """
    cotangent: Tensor

    @classmethod
    def create(cls: Type[T], shape: Tuple[int, ...]) -> T:
        """
        Factory to create LogCotangent object.__class__(
        Args:
            shape: The shape of the transmitted tensor and its cotangent.
        """
        return cls(jnp.zeros(shape))

    def forward(self, x: Tensor) -> Tensor:
        """
        This method is called in the forward pass.  It will automatically log the cotangent in the
        backward pass.

        Args:
            x: The tensor to be transmitted.
        Returns: The same tensor that was inputted, x.
        """
        if x.shape != self.cotangent.shape:
            raise ValueError
        return x + self.cotangent



from __future__ import annotations

from functools import partial
from typing import Generic, Tuple

from jax import numpy as jnp
from jax import vjp
from jax.tree_util import tree_multimap

from tjax import custom_vjp, dataclass

from .augmented import State
from .comparing import ComparingIteratedFunction
from .iterated_function import IteratedFunction, Parameters, TheAugmentedState

__all__ = ['IteratedFunctionWithCombinator']


@dataclass
class _ZResiduals(Generic[Parameters, State, TheAugmentedState]):
    outer_iterated_function: IteratedFunctionWithCombinator[Parameters, State, TheAugmentedState]
    outer_theta: Parameters
    x_star: State


@dataclass
class _ZParameters(Generic[Parameters, State]):
    outer_theta: Parameters
    x_star: State
    x_star_bar: State


@dataclass
class _ZIterate(ComparingIteratedFunction[_ZParameters[Parameters, State], State, State],
                Generic[Parameters, State, TheAugmentedState]):

    iterated_function: IteratedFunction[Parameters, State, TheAugmentedState]

    # Implemented methods --------------------------------------------------------------------------
    def iterate_state(self, theta: _ZParameters[Parameters, State], x: State) -> State:
        # The state should be called z, but we can't change the interface because of Liskov's
        # substitution principle.
        z = x
        del x

        def f_of_x(x: State) -> State:
            return self.iterated_function.iterate_state(theta.outer_theta, x)

        _, df_by_dx = vjp(f_of_x, theta.x_star)
        df_by_dx_times_z, = df_by_dx(z)
        return tree_multimap(jnp.add, theta.x_star_bar, df_by_dx_times_z)

    def extract_comparand(self, state: State) -> State:
        return state


def _ffp_fwd(outer_iterated_function: IteratedFunctionWithCombinator[Parameters, State,
                                                                     TheAugmentedState],
             theta: Parameters,
             initial_state: State) -> Tuple[TheAugmentedState, _ZResiduals[Parameters, State,
                                                                           TheAugmentedState]]:
    """
    Args:
        theta: The parameters for which gradients can be calculated.
        initial_state: An initial guess of the final state.
    Returns:
        x_star: the result of the minimization.
        residuals: residuals used in _ffp_bwd.
    """
    augmented: TheAugmentedState = outer_iterated_function.find_fixed_point(
        theta, initial_state)
    return augmented, _ZResiduals(outer_iterated_function, theta, augmented.current_state)


def _ffp_bwd(residuals: _ZResiduals[Parameters, State, TheAugmentedState],
             augmented_star_bar: TheAugmentedState) -> Tuple[Parameters]:
    """
    Args:
        residuals: residuals produced by _ffp_fwd.
        augmented_star_bar: cotangents
    Returns:
        theta_bar: cotangents for theta
        zeroed_xs: cotangents for initial_state
    """
    # pylint: disable=protected-access
    outer_iterated_function = residuals.outer_iterated_function
    x_star_bar = augmented_star_bar.current_state

    def f_of_theta(some_theta: Parameters) -> State:
        return outer_iterated_function.iterate_state(some_theta, residuals.x_star)

    z_iterator = _ZIterate(iteration_limit=outer_iterated_function.z_iteration_limit,
                           iterated_function=outer_iterated_function)
    z_parameters = _ZParameters(residuals.outer_theta, residuals.x_star, x_star_bar)
    z_star: State = z_iterator.find_fixed_point(z_parameters, x_star_bar).current_state

    _, df_by_dtheta = vjp(f_of_theta, residuals.outer_theta)
    theta_bar, = df_by_dtheta(z_star)
    return (theta_bar,)


# https://github.com/python/mypy/issues/8539
@dataclass  # type: ignore
class IteratedFunctionWithCombinator(IteratedFunction[Parameters, State, TheAugmentedState],
                                     Generic[Parameters, State, TheAugmentedState]):
    """
    An IteratedFunctionWithCombinator is an IteratedFunction that invokes a combinator so that
    differentiation works through the fixed point.  Besides inheriting from this class, no other
    action is necessary to get this capability.

    Attributes:
        z_iteration_limit:
            The maximum number of iterations to use to evaluate the adjoint's fixed point.
    """
    z_iteration_limit: int = 1000

    # Overridden methods ---------------------------------------------------------------------------
    @partial(custom_vjp[TheAugmentedState], nondiff_argnums=(0, 2))
    def find_fixed_point(self,  # type: ignore
                         theta: Parameters,
                         initial_state: State) -> TheAugmentedState:
        """
        Args:
            theta: The parameters for which gradients can be calculated.
            initial_state: An initial guess of the final state.
        Returns: The augmented state at the fixed point.
        """
        return super().find_fixed_point(theta, initial_state)

    # Apply vjp ------------------------------------------------------------------------------------
    find_fixed_point.defvjp(_ffp_fwd, _ffp_bwd)



from __future__ import annotations

from functools import partial
from typing import Any, Callable, Generic, Optional, Tuple, TypeVar

from jax import jit
from jax import numpy as jnp
from jax.experimental.host_callback import id_tap
from jax.lax import scan, while_loop
from jax.tree_util import tree_map, tree_multimap

from tjax import PyTree, Tensor, dataclass, default_atol, default_rtol

from .augmented import AugmentedState, State

__all__ = ['IteratedFunction']


Parameters = TypeVar('Parameters', bound=PyTree)
Trajectory = TypeVar('Trajectory', bound=PyTree)
TheAugmentedState = TypeVar('TheAugmentedState', bound=AugmentedState[Any])


# https://github.com/python/mypy/issues/8539
@dataclass  # type: ignore
class IteratedFunction(Generic[Parameters, State, TheAugmentedState]):
    """
    An IteratedFunction object models an iterated function.

    It is a generic class in terms of two generic types:
        * Parameters, which models a PyTree subtype for parameters, and
        * State, which models a PyTree subtype for the state.

    The minimum requirement is to implement iterate_state, which iterates the state given
    parameters.

    In the case of a stochastic iterated function, iterate_augmented should be augmented, and should
    return the next augmented state.  iterate_state should return the expected value of the next
    stochastic state.

    It is possible to modify the stopping condition by overriding state_needs_iteration.

    The two main methods of IteratedFunction are
    * find_fixed_point, which iterates until the fixed point is reached or other stopping conditions
      are met, and
    * sample_trajectory, which iterates for a fixed number of iterations and returns a trajectory.

    Args:
        iteration_limit: The maximum number of iterations for this fixed point solver.  This must be
            positive.
        rtol: The relative tolerance for the comparison stopping condition.
        atol: The absolute tolerance for the comparison stopping condition.
        initial_rng: The inital random number generator.
    """

    minimum_iterations: int = 10
    iteration_limit: int
    rtol: float = default_rtol
    atol: float = default_atol

    # New methods ----------------------------------------------------------------------------------
    @jit
    def find_fixed_point(self, theta: Parameters, initial_state: State) -> TheAugmentedState:
        """
        Args:
            theta: The parameters for which gradients can be calculated.
            initial_state: An initial guess of the final state.
        Returns:
            state: The augmented state at the fixed point.
        """
        return while_loop(partial(self.state_needs_iteration, theta),
                          partial(self.iterate_augmented, theta),
                          self.initial_state(initial_state))

    @partial(jit, static_argnums=(3, 4, 5))
    def sample_trajectory(self,
                          theta: Parameters,
                          initial_state: State,
                          iteration_limit: int,
                          tap_function: Optional[Callable[[None], None]],
                          extract: Callable[[TheAugmentedState], Trajectory]) -> (
                              Tuple[TheAugmentedState, Trajectory]):
        """
        Args:
            theta: The parameters for which gradients can be calculated.
            initial_state: An initial guess of the final state.
            iteration_limit: The number of steps in the trajectory.  Unlike the eponymous member
                variable, this must be static.
        Returns:
            x_star: The augmented state at the fixed point.
            trajectory: A PyTree representing the trajectory of states.
        """
        def f(augmented: TheAugmentedState, x: None) -> Tuple[
                TheAugmentedState, Trajectory]:
            new_augmented = self.iterate_augmented(theta, augmented)
            extracted = extract(new_augmented)
            if tap_function is not None:
                extracted = id_tap(tap_function, None, result=extracted)
            return new_augmented, extracted
        return scan(f, self.initial_state(initial_state), None, iteration_limit)

    def debug_fixed_point(self, theta: Parameters, initial_state: State) -> TheAugmentedState:
        """
        This method is identical to find_fixed_point, but avoids using while_loop and its
        concomitant jit.
        """
        augmented = self.initial_state(initial_state)
        while self.state_needs_iteration(theta, augmented):
            augmented = self.iterate_augmented(theta, augmented)
        return augmented

    def debug_trajectory(self,
                         theta: Parameters,
                         initial_state: State,
                         iteration_limit: int,
                         tap_function: Optional[Callable[[None], None]],
                         extract: Callable[[TheAugmentedState], Trajectory]) -> (
                             Tuple[TheAugmentedState, Trajectory]):
        """
        This method is identical to sample_trajectory, but avoids using scan and its concomitant
        jit.
        """
        augmented = self.initial_state(initial_state)
        trajectory = tree_map(lambda x: jnp.zeros((0,) + x.shape), extract(augmented))
        for _ in range(iteration_limit):
            augmented = self.iterate_augmented(theta, augmented)
            extracted = extract(augmented)
            trajectory = tree_multimap(jnp.append, trajectory, extracted)
            if tap_function is not None:
                tap_function(None)
        return augmented, trajectory

    def state_needs_iteration(self, theta: Parameters, augmented: TheAugmentedState) -> bool:
        """
        Args:
            theta: The parameters.
            augmented: The state.
        Returns: True while iteration needs to continue.
        """
        converged = jnp.logical_not(jnp.logical_and(augmented.iterations > self.minimum_iterations,
                                                    self.converged(augmented)))
        too_many_iterations = augmented.iterations < self.iteration_limit
        return jnp.logical_and(too_many_iterations, converged)

    # Abstract methods -----------------------------------------------------------------------------
    def initial_state(self, initial_state: State) -> TheAugmentedState:
        raise NotImplementedError

    def iterate_state(self, theta: Parameters, x: State) -> State:
        """
        Returns: The expected value of the next state given the old one.  A state must be
            differentiable.
        """
        raise NotImplementedError

    def iterate_augmented(self,
                          theta: Parameters,
                          augmented: TheAugmentedState) -> TheAugmentedState:
        """
        Returns: The value of the next augmented state given the old one.
        """
        raise NotImplementedError

    def converged(self, augmented: TheAugmentedState) -> Tensor:
        """
        Returns: A Boolean Tensor of shape () indicating whether the pytrees are close.
        """
        raise NotImplementedError



from __future__ import annotations

from typing import Any, List, Optional, Tuple

import jax.random
from jax import numpy as jnp

from .annotations import Shape, Tensor
from .dataclass import dataclass

__all__ = ['Generator']


@dataclass
class Generator:
    """
    This class represents a JAX random number generator.  Unlike `numpy.Generator`, `tjax.Generator`
    has no mutating methods.  Instead, its generation methods return a new instance along with
    the generated tensor.
    """

    key: Tensor

    def __init__(self,
                 *,
                 seed: Optional[int] = None,
                 key: Optional[Tensor] = None,
                 **kwargs: Any):
        super().__init__(**kwargs)
        if key is None:
            if seed is None:
                raise ValueError
            key = jax.random.PRNGKey(seed)
        object.__setattr__(self, 'key', key)

    # New methods ----------------------------------------------------------------------------------
    def split(self, n: int = 2) -> List[Generator]:
        keys = jax.random.split(self.key, n)
        return [Generator(key=key) for key in keys]

    def normal(self, std_dev: Tensor, shape: Shape = ()) -> (
            Tuple[Generator, Tensor]):
        g1, g2 = self.split()
        return g1, std_dev * jax.random.normal(g2.key, shape)

    def gamma(self, gamma_shape: Tensor, shape: Shape = ()) -> (
            Tuple[Generator, Tensor]):
        g1, g2 = self.split()
        return g1, jax.random.gamma(g2.key, gamma_shape, shape)

    # Magic methods --------------------------------------------------------------------------------
    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Generator):
            return NotImplemented
        return jnp.all(self.key == other.key)

    def __ne__(self, other: Any) -> bool:
        if not isinstance(other, Generator):
            return NotImplemented
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash((int(self.key[0]), int(self.key[1])))



from functools import reduce
from numbers import Number
from operator import add
from typing import Any, Collection, Optional

import numpy as np
from jax import numpy as jnp

from .annotations import ShapeLike, Tensor

__all__ = ['sum_tensors', 'is_scalar']


def sum_tensors(tensors: Collection[Tensor],
                shape: Optional[ShapeLike] = None) -> Tensor:
    if not tensors:
        return jnp.zeros(shape)
    return reduce(add, tensors)


def is_scalar(x: Any) -> bool:
    return isinstance(x, Number) or isinstance(x, (np.ndarray, jnp.ndarray)) and x.shape == ()



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, Hashable, List, Sequence, Tuple, Union

import numpy as np
from jax import numpy as jnp

__all__ = ['Shape',
           'ShapeLike',
           'SliceLike',
           'Tensor',
           'RealTensor',
           'ComplexTensor',
           'PyTree']


Shape = Tuple[int, ...]
ShapeLike = Union[int, Sequence[int]]
SliceLike = Tuple[Union[int, None, slice], ...]
Tensor = Union[np.ndarray, jnp.ndarray]
RealTensor = Tensor
ComplexTensor = Tensor
PyTree = Union[Tensor,
               'PyTreeLike',
               Tuple['PyTree', ...],
               List['PyTree'],
               Dict[Hashable, 'PyTree'],
               None]


if TYPE_CHECKING:
    from .pytree_like import PyTreeLike



"""
This library implements a variety of tools for the differential programming library
[JAX](https://github.com/google/jax).
"""
from .annotations import *
from .color_stub import *
from .dataclass import *
from .display import *
from .dtypes import *
from .generator import *
from .graph import *
from .leaky_integral import *
from .log_cotangent import *
from .partial import *
from .pytree_like import *
from .shims import *
from .testing import *
from .tools import *

__pdoc__ = {}
__pdoc__['real_dtype'] = False
__pdoc__['complex_dtype'] = False
__pdoc__['PyTreeLike'] = False
__pdoc__['Field'] = False
__pdoc__['InitVar'] = False
__pdoc__['FrozenInstanceError'] = False
from .dataclass import document_dataclass

document_dataclass(__pdoc__, 'Generator')
document_dataclass(__pdoc__, 'Partial')
document_dataclass(__pdoc__, 'LogCotangent')
del document_dataclass


__all__ = list(locals())



import sys
from typing import List

import colorful as cf

__all__: List[str] = []


cf.use_style('solarized')
if not sys.stdout.isatty() or not sys.stderr.isatty():
    # Output is piped or redirected.
    cf.disable()



from jax import numpy as jnp
from jax.dtypes import canonicalize_dtype

__all__ = ['int_dtype',
           'real_dtype',
           'complex_dtype',
           'default_rtol',
           'default_atol',
           'default_tols']


# This may one day not require canonicalize_dtype.  Check if jnp.float_ is still np.float64.  One of
# the downsides of calling canonicalize_dtype here is that this file must be imported after
# config.update("jax_enable_x64", True) or else canonicalize_dtype will already have cached the
# default float width.
int_dtype = canonicalize_dtype(jnp.int_).type
"The dtype used by JAX for int values.  Typically, either `numpy.int32` or `numpy.int64`."

real_dtype = canonicalize_dtype(jnp.float_).type
"The dtype used by JAX for real values.  Typically, either `numpy.float32` or `numpy.float64`."

complex_dtype = canonicalize_dtype(jnp.complex_).type
"""
The dtype used by JAX for complex values.  Typically, either `numpy.complex64` or
`numpy.complex128`.
"""


default_rtol = {jnp.float32: 1e-4,
                jnp.float64: 1e-5}[real_dtype]
default_atol = {jnp.float32: 1e-6,
                jnp.float64: 1e-8}[real_dtype]
default_tols = dict(rtol=default_rtol, atol=default_atol)



from typing import Any, Callable, Generic, Tuple, TypeVar, Union

import jax
from jax import numpy as jnp
from jax.tree_util import Partial, tree_map

__all__ = ['jit', 'custom_jvp', 'custom_vjp']


R = TypeVar('R')
F = TypeVar('F', bound=Callable[..., Any])


def as_sorted_tuple(x: Union[int, Tuple[int, ...]]) -> Tuple[int, ...]:
    if isinstance(x, int):
        return (x,)
    return tuple(sorted(x))


def jit(func: F, **kwargs: Any) -> F:
    retval = jax.jit(func, **kwargs)
    if hasattr(func, "__isabstractmethod__"):
        retval.__isabstractmethod__ = func.__isabstractmethod__  # type: ignore
    return retval


class custom_vjp(Generic[R]):
    """
    Set up a JAX-transformable function for a custom VJP rule definition.

    This class is meant to be used as a function decorator. Instances are callables that behave
    similarly to the underlying function to which the decorator was applied, except when a
    reverse-mode differentiation transformation (like `jax.grad()`) is applied, in which case a
    custom user-supplied VJP rule function is used instead of tracing into and performing automatic
    differentiation of the underlying function’s implementation. There is a single instance method,
    defvjp, which defines the custom VJP rule.

    This decorator precludes the use of forward-mode automatic differentiation.

    This is a shim class to work around an
    [issue with JAX's custom_vjp](https://github.com/google/jax/issues/2912).  It provides both:

    - static arguments, and
    - nondifferentiable arguments.

    Static arguments are passed in to both the forward and the backward pass.  They must be
    hashable.  Different values for static arguments will generate recompilation.

    The generated backward pass will generate zeroed-out cotangents.  Ideally, no corresponding
    cotangents would be created, but such a change would have to be done in JAX itself.

    For example::

        from tjax import custom_vjp
        from jax import numpy as jnp

        @partial(custom_vjp, nondiff_argnums=2)
        def f(x, y, z):
        return jnp.sin(x) * y + z

        def f_fwd(x, y, z):
        return f(x, y, z), (jnp.cos(x), jnp.sin(x), y)

        def f_bwd(residuals, output_bar):
        cos_x, sin_x, y = residuals
        x_bar = cos_x * output_bar * y
        y_bar = sin_x * output_bar
        # z_bar is not returned because it's nondifferentiable.
        return x_bar, y_bar

        f.defvjp(f_fwd, f_bwd)
    """
    def __init__(self,
                 fun: Callable[..., R],
                 static_argnums: Union[int, Tuple[int, ...]] = (),
                 nondiff_argnums: Union[int, Tuple[int, ...]] = ()):
        """
        Args:
            fun: the function to decorate.
            static_argnums: The indices of the static arguments.
            nondiff_argnums: The indices of the nondifferentiable arguments.
        """
        static_argnums = as_sorted_tuple(static_argnums)
        nondiff_argnums = as_sorted_tuple(nondiff_argnums)
        intersection = set(static_argnums) & set(nondiff_argnums)
        if intersection:
            raise ValueError(
                f"Arguments {intersection} cannot be both static and nondifferentiable.")
        self.nondiff_argnums = nondiff_argnums
        self.vjp = jax.custom_vjp(fun, nondiff_argnums=static_argnums)

    def defvjp(self, fwd: Callable[..., Tuple[R, Any]], bwd: Callable[..., Any]) -> None:
        """
        Implement the custom forward and backward passes of the custom derivative.

        Args:
            fwd: The custom forward pass.
            bwd: The custom backward pass.  Cotangents for the nondifferentiable arguments should
                not be provided by the user-provided backward pass.
        """
        def new_fwd(*args: Any) -> Tuple[R, Any]:
            zeroed_args = tuple([tree_map(jnp.zeros_like, args[i])
                                 for i in self.nondiff_argnums])
            primal, internal_residuals = fwd(*args)
            return primal, (zeroed_args, internal_residuals)

        def new_bwd(residuals: Any, output_bar: R) -> Any:
            zeroed_args, internal_residuals = residuals
            input_bar = bwd(internal_residuals, output_bar)
            input_bar = list(input_bar)
            for i, index in enumerate(self.nondiff_argnums):
                input_bar[index: index] = [zeroed_args[i]]
            return tuple(input_bar)

        self.vjp.defvjp(new_fwd, new_bwd)

    def __call__(self, *args: Any) -> R:
        return self.vjp(*args)

    def __get__(self, instance: Any, owner: Any = None) -> Callable[..., R]:
        if instance is None:
            return self
        # Create a partial function application corresponding to a bound method.
        return Partial(self, instance)


class custom_jvp(jax.custom_jvp, Generic[R]):
    """Set up a JAX-transformable function for a custom JVP rule definition.

    This class is meant to be used as a function decorator. Instances are
    callables that behave similarly to the underlying function to which the
    decorator was applied, except when a differentiation transformation (like
    :py:func:`jax.jvp` or :py:func:`jax.grad`) is applied, in which case a custom user-supplied
    JVP rule function is used instead of tracing into and performing automatic
    differentiation of the underlying function's implementation. There is a single
    instance method, ``defjvp``, which defines the custom JVP rule.

    For example::

    import jax.numpy as jnp

    @jax.custom_jvp
    def f(x, y):
        return jnp.sin(x) * y

    @f.defjvp
    def f_jvp(primals, tangents):
        x, y = primals
        x_dot, y_dot = tangents
        primal_out = f(x, y)
        tangent_out = jnp.cos(x) * x_dot * y + jnp.sin(x) * y_dot
        return primal_out, tangent_out

    For a more detailed introduction, see the tutorial_.
    """
    def __init__(self,
                 fun: Callable[..., R],
                 static_argnums: Union[int, Tuple[int, ...]] = (),
                 nondiff_argnums: Union[int, Tuple[int, ...]] = ()):
        """
        Args:
            fun: the function to decorate.
            static_argnums: The indices of the static arguments.
            nondiff_argnums: The indices of the nondifferentiable arguments.
        """
        static_argnums = as_sorted_tuple(static_argnums)
        nondiff_argnums = as_sorted_tuple(nondiff_argnums)
        intersection = set(static_argnums) & set(nondiff_argnums)
        if intersection:
            raise ValueError(
                f"Arguments {intersection} cannot be both static and nondifferentiable.")
        self.nondiff_argnums_excluding_static_args = [
            nondiff_argnum - sum(1 for x in static_argnums if x < nondiff_argnum)
            for nondiff_argnum in nondiff_argnums]
        # self.len_static_argnums = len(static_argnums)
        self.nondiff_argnums = nondiff_argnums
        self.jvp = jax.custom_jvp(fun, nondiff_argnums=static_argnums)

    def defjvp(self, jvp: Callable[..., Any]) -> None:
        """
        Implement the custom forward pass of the custom derivative.

        Args:
            fwd: The custom forward pass.
        """
        def new_fwd(*args: Any) -> Tuple[R, Any]:
            static_args = args[: -2]
            primals, tangents = args[-2:]
            nondiff_args = [primals[i] for i in self.nondiff_argnums_excluding_static_args]
            new_primals = [p
                           for i, p in enumerate(primals)
                           if i not in self.nondiff_argnums_excluding_static_args]
            new_tangents = [t
                            for i, t in enumerate(tangents)
                            if i not in self.nondiff_argnums_excluding_static_args]
            return self.jvp(*static_args, *nondiff_args, new_primals, new_tangents)

        self.jvp.defjvp(new_fwd)

    def __call__(self, *args: Any) -> R:
        return self.jvp(*args)

    def __get__(self, instance: Any, owner: Any = None) -> Callable[..., R]:
        if instance is None:
            return self
        # Create a partial function application corresponding to a bound method.
        return Partial(self, instance)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tjax', 'tjax.fixed_point']

package_data = \
{'': ['*']}

install_requires = \
['colorful>=0.5.4,<0.6.0',
 'cooperative_dataclasses>=0.2.0,<0.3.0',
 'jax>=0.1.72,<0.2.0',
 'jaxlib>=0.1.51,<0.2.0',
 'networkx>=2.4,<3.0',
 'numpy>=1.19,<2.0']

setup_kwargs = {
    'name': 'tjax',
    'version': '0.3.10',
    'description': 'Tools for JAX.',
    'long_description': '=============\nTools for JAX\n=============\n\n.. role:: bash(code)\n    :language: bash\n\n.. role:: python(code)\n   :language: python\n\nThis repository implements a variety of tools for the differential programming library\n`JAX <https://github.com/google/jax>`_.  It includes:\n\n- A dataclass decorator that facilitates defining JAX trees, provides convenient text display, and\n  provides a mypy plugin\n\n- A custom VJP decorator that supports both static and non-differentiable arguments\n\n- A random number generator class\n\n- JAX tree registration for `NetworkX <https://networkx.github.io/>`_ graph types\n\n- Testing tools that automatically produce testing code\n\nSee the `documentation <https://neilgirdhar.github.io/tjax/tjax/index.html>`_.\n\nContribution guidelines\n=======================\n\n- Conventions: PEP8.\n\n- How to run tests: :bash:`pytest .`\n\n- How to clean the source:\n\n  - :bash:`isort tjax`\n  - :bash:`pylint tjax`\n  - :bash:`mypy tjax`\n  - :bash:`flake8 tjax`\n',
    'author': 'Neil Girdhar',
    'author_email': 'mistersheik@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/NeilGirdhar/cmm',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from functools import partial
from typing import Generic, Tuple, TypeVar

from jax import numpy as jnp
from jax.tree_util import tree_map, tree_multimap, tree_reduce

from tjax import Generator, PyTree, Tensor, dataclass

from ..leaky_integral import leaky_integrate
from .augmented import AugmentedState, State
from .combinator import IteratedFunctionWithCombinator
from .iterated_function import IteratedFunction, Parameters

__all__ = ['StochasticState', 'StochasticIteratedFunction',
           'StochasticIteratedFunctionWithCombinator']


Comparand = TypeVar('Comparand', bound=PyTree)


@dataclass
class StochasticState(AugmentedState[State], Generic[State, Comparand]):

    mean_state: Comparand
    second_moment_state: Comparand
    rng: Generator


# https://github.com/python/mypy/issues/8539
@dataclass  # type: ignore
class StochasticIteratedFunction(
        IteratedFunction[Parameters, State, StochasticState[State, Comparand]],
        Generic[Parameters, State, Comparand]):

    initial_rng: Generator
    decay: float

    # Implemented methods --------------------------------------------------------------------------
    def initial_state(self, initial_state: State) -> StochasticState[State, Comparand]:
        mean_state, second_moment_state = self._sufficient_statistics(initial_state)
        return StochasticState(current_state=initial_state,
                               iterations=0,
                               mean_state=mean_state,
                               second_moment_state=second_moment_state,
                               rng=self.initial_rng)

    def iterate_augmented(self,
                          theta: Parameters,
                          augmented: StochasticState[State, Comparand]) -> (
                              StochasticState[State, Comparand]):
        def f(value: Tensor, drift: Tensor) -> Tensor:
            return leaky_integrate(value, 1.0, drift, self.decay, leaky_average=True)

        new_state, new_rng = self.stochastic_iterate_state(
            theta, augmented.current_state, augmented.rng)
        mean_state, second_moment_state = self._sufficient_statistics(augmented.current_state)
        new_mean_state = tree_multimap(f, augmented.mean_state, mean_state)
        new_second_moment_state = tree_multimap(f, augmented.second_moment_state,
                                                second_moment_state)
        return StochasticState(current_state=new_state,
                               iterations=augmented.iterations + 1,
                               mean_state=new_mean_state,
                               second_moment_state=new_second_moment_state,
                               rng=new_rng)

    def converged(self, augmented: StochasticState[State, Comparand]) -> Tensor:
        data_weight = leaky_integrate(0.0, augmented.iterations, 1.0, self.decay,
                                      leaky_average=True)
        mean_squared = tree_map(jnp.square, augmented.mean_state)
        return tree_reduce(jnp.logical_and,
                           tree_multimap(partial(jnp.allclose,
                                                 rtol=self.rtol * data_weight,
                                                 atol=self.atol * data_weight),
                                         augmented.second_moment_state,
                                         mean_squared))

    # Abstract methods -----------------------------------------------------------------------------
    def extract_comparand(self, state: State) -> Comparand:
        """
        Returns: A pytree that will be compared in successive states to check whether the state has
            converged.
        """
        raise NotImplementedError

    def stochastic_iterate_state(self,
                                 theta: Parameters,
                                 state: State,
                                 rng: Generator) -> Tuple[State, Generator]:
        raise NotImplementedError

    # New methods ----------------------------------------------------------------------------------
    def convergence_atol(self, augmented: StochasticState[State, Comparand]) -> Tensor:
        data_weight = leaky_integrate(0.0, augmented.iterations, 1.0, self.decay,
                                      leaky_average=True)
        mean_squared = tree_map(jnp.square, augmented.mean_state)
        variance = tree_multimap(jnp.subtract, augmented.second_moment_state, mean_squared)
        return tree_reduce(jnp.maximum, tree_map(jnp.amax, variance)) / data_weight

    # Private methods ------------------------------------------------------------------------------
    def _sufficient_statistics(self, state: State) -> Tuple[Comparand, Comparand]:
        comparand = self.extract_comparand(state)
        squared_comparand = tree_map(jnp.square, comparand)
        return comparand, squared_comparand


class StochasticIteratedFunctionWithCombinator(
        IteratedFunctionWithCombinator[Parameters, State, StochasticState[State, Comparand]],
        StochasticIteratedFunction[Parameters, State, Comparand],
        Generic[Parameters, State, Comparand]):
    pass



from __future__ import annotations

from functools import partial
from typing import Generic, TypeVar

from jax import numpy as jnp
from jax.tree_util import tree_multimap, tree_reduce

from tjax import PyTree, Tensor, dataclass

from .augmented import AugmentedState, State
from .iterated_function import IteratedFunction, Parameters

__all__ = ['ComparingState', 'ComparingIteratedFunction']


Comparand = TypeVar('Comparand', bound=PyTree)


@dataclass
class ComparingState(AugmentedState[State], Generic[State, Comparand]):

    last_state: Comparand


# https://github.com/python/mypy/issues/8539
@dataclass  # type: ignore
class ComparingIteratedFunction(
        IteratedFunction[Parameters, State, ComparingState[State, Comparand]],
        Generic[Parameters, State, Comparand]):

    # Implemented methods --------------------------------------------------------------------------
    def initial_state(self, initial_state: State) -> ComparingState[State, Comparand]:
        return ComparingState(current_state=initial_state,
                              iterations=0,
                              last_state=self.extract_comparand(initial_state))

    def iterate_augmented(self,
                          theta: Parameters,
                          augmented: ComparingState[State, Comparand]) -> (
                              ComparingState[State, Comparand]):
        return ComparingState(current_state=self.iterate_state(theta, augmented.current_state),
                              iterations=augmented.iterations + 1,
                              last_state=self.extract_comparand(augmented.current_state))

    def converged(self, augmented: ComparingState[State, Comparand]) -> Tensor:
        return tree_reduce(jnp.logical_and,
                           tree_multimap(partial(jnp.allclose, rtol=self.rtol, atol=self.atol),
                                         self.extract_comparand(augmented.current_state),
                                         augmented.last_state))

    # New methods ----------------------------------------------------------------------------------
    def extract_comparand(self, state: State) -> Comparand:
        """
        Returns: A pytree that will be compared in successive states to check whether the state has
            converged.
        """
        raise NotImplementedError



import sys
from numbers import Number
from typing import Any, Type, Union

import colorful as cf
import networkx as nx
import numpy as np
from jax.interpreters.ad import JVPTracer
from jax.interpreters.partial_eval import JaxprTracer
from jax.interpreters.xla import DeviceArray

__all__ = ['Displayable', 'print_generic']


if sys.version_info >= (3, 8):
    from typing import Protocol, runtime_checkable

    @runtime_checkable
    class Displayable(Protocol):
        """
        This protocol identifies classes that support the `display_generic` mechanism.
        """
        def display(self, show_values: bool = True, indent: int = 0) -> str:
            ...
else:
    class Displayable:
        pass


def print_generic(*args: Any, **kwargs: Any) -> None:
    for value in args:
        print(display_generic(value))
    for key, value in kwargs.items():
        print(display_key_and_value(key, value, "=", True, 0))


def display_generic(value: Any, show_values: bool = True, indent: int = 0) -> str:
    # pylint: disable=too-many-return-statements
    if isinstance(value, JVPTracer):
        return cf.magenta(f"JVPTracer {value.shape} {value.dtype}") + "\n"
    if isinstance(value, JaxprTracer):
        return cf.magenta(f"JaxprTracer {value.shape} {value.dtype}") + "\n"
    if isinstance(value, (np.ndarray, DeviceArray)):
        if isinstance(value, np.ndarray):
            retval = cf.yellow(f"NumPy Array {value.shape}") + "\n"
        elif isinstance(value, DeviceArray):
            retval = cf.violet(f"Jax Array {value.shape}") + "\n"
        if show_values:
            retval += _show_array(indent + 1, value)
        return retval
    if isinstance(value, Number) or value is None:
        return cf.cyan(str(value)) + "\n"
    if (isinstance(value, Displayable)
            if sys.version_info >= (3, 8)
            else hasattr(value, 'display')):
        return value.display(show_values=show_values, indent=indent)
    if isinstance(value, dict):
        return (display_class(dict)
                + "".join(display_key_and_value(key, sub_value, "=", show_values, indent)
                          for key, sub_value in value.items()))
    if isinstance(value, (tuple, list)):
        return (display_class(type(value))
                + "".join(display_key_and_value("", sub_value, "", show_values, indent)
                          for sub_value in value))
    if isinstance(value, nx.Graph):
        return display_graph(value, show_values, indent)
    return cf.red(str(value)) + "\n"


def display_graph(graph: nx.Graph, show_values: bool, indent: int = 0) -> str:
    directed = isinstance(graph, nx.DiGraph)
    arrow = cf.base00('⟶  ' if directed else '🡘 ')
    retval = display_class(type(graph))
    for name, value in graph.nodes.items():
        retval += display_key_and_value(name, value['element'], ": ", show_values, indent)
    for (source, target), value in graph.edges.items():
        key = f"{source}{arrow}{target}"
        retval += display_key_and_value(key, value['link'], ": ", show_values, indent)
    return retval


def display_class(cls: Type[Any]) -> str:
    return cf.orange(cls.__name__) + "\n"


def display_key_and_value(
        key: str, value: Any, separator: str, show_values: bool = True, indent: int = 0) -> str:
    return (_indent_space(indent + 1) + cf.blue(key) + cf.base00(separator)
            + display_generic(value, show_values, indent + 1))


# Private functions --------------------------------------------------------------------------------
_indentation = 4


def _indent_space(indent: int) -> str:
    return (indent * _indentation) * " "


def _format_number(x: float) -> str:
    return f"{x:10.4f}"


def _show_array(indent: int, array: Union[np.ndarray, DeviceArray]) -> str:
    if len(array.shape) == 0:
        return _indent_space(indent) + _format_number(array[()]) + "\n"
    if np.prod(array.shape) == 0:
        return ""
    if 1 in array.shape:
        return _show_array(indent, array[tuple(0 if s == 1 else slice(None)
                                               for s in array.shape)])
    if len(array.shape) == 1:
        return _indent_space(indent) + "  ".join(_format_number(array[i])
                                                 for i in range(array.shape[0])) + "\n"
    if len(array.shape) == 2:
        return "".join(_show_array(indent, array_slice)
                       for array_slice in array)
    return ""
